# HACC-FFT1D

https://git.cels.anl.gov/hacc/hacc-fft1d

## Goal

hacc-fft1d replicates the shared-memory 1D FFT batches that each MPI rank needs to compute as part of HACC's 3D global FFT, which is the core of HACC's global Poisson solver. For each global 3D FFT, each rank would need to compute 3 batches of 1D FFTs (x, y, and z directions). Each of HACC's global Poisson solve operations executes 1 forward and 3 backward (for gradient components) global 3D FFTs. So each Poisson solve executes 3 batches of 1D forward FFTs and 9 batches of 1D backward FFTs, and there are generally 500-1000 Poisson solves per high resolution HACC simulation.

## Building

hacc-fft1d is implemented in C++, uses the FFTW3 API, and can use OpenMP3 threads. An example configuration of FFTW3 sufficient to run hacc-fft1d is:
```
CC=icc ./configure --enable-openmp --disable-mpi --disable-fortran --prefix=/path/to/install
```

hacc-fft1d includes an example `makefile` demonstrating how to build against FFTW3. It could be modified to link against a vendor FFT library that implements the FFTW3 API. The code could also be modified to use another FFT API that performs the equivalent of FFTW3's `fftw_plan_many_dft()`.

## Running

Running hacc-fft1d without command line arguments shows a summary of usage info:
```
USAGE: ./hacc-fft1d <size> <howmany> <repetitions>
```
`size` is the length of each 1D FFT, `howmany` is the number of 1D FFTs in a batch for each MPI rank, and `repetitions` is the number of times to sequentially run the test. Each test starts with a Dirac delta in each 1D FFT and performs a forward and backward FFT. Many implementations conforming to the FFTW3 API are able to produce bitwise consistent results for the Dirac delta test, so hacc-fft1d checks every element of every 1D array both in the transform space and back to the original space and reports the total number of elements that do not have the expected value. hacc-fft1d reports the wall time in seconds required to build plans, setup arrays, perform the 1D FFT batches, and check results. hacc-fft1d runs a sequence of tests for plans build with `FFTW_ESTIMATE` and then repeats the sequence of tests for plans build with `FFTW_MEASURE`. In practice the wall time taken to build the `FFTW_MEASURE` plans is slightly less after having built plans with `FFTW_ESTIMATE` than when starting without having generated any plans, and hacc-fft1d could be slightly modified to make the timing of planning with `FFTW_MEASURE` more accurate.

Here is an example output from running hacc_fft1d on some sort of Xeon:
```
OMP_NUM_THREADS=8 ./hacc-fft1d 38400 12288 2

omp_max_threads = 8
fftw_planner_nthreads = 8


FFTW_ESTIMATE
t_plan = 0.000972 seconds

0
t_setup = 1.066796 seconds
t_fftf = 1.961796 seconds
t_check = 0.225370 seconds
n_errors = 0
t_fftb = 2.500236 seconds
t_check = 0.161093 seconds
n_errors = 0

1
t_setup = 0.675114 seconds
t_fftf = 1.673704 seconds
t_check = 0.137304 seconds
n_errors = 0
t_fftb = 1.274783 seconds
t_check = 0.235393 seconds
n_errors = 0


FFTW_MEASURE
t_plan = 2.210040 seconds

0
t_setup = 0.318225 seconds
t_fftf = 1.257030 seconds
t_check = 0.132910 seconds
n_errors = 0
t_fftb = 0.883046 seconds
t_check = 0.093539 seconds
n_errors = 0

1
t_setup = 0.174378 seconds
t_fftf = 0.785495 seconds
t_check = 0.092930 seconds
n_errors = 0
t_fftb = 0.786101 seconds
t_check = 0.094062 seconds
n_errors = 0

```

## Exascale Problem Sizes

Assuming an exascale system with 10,000 nodes and 12 MPI ranks per node, here are some ballpark problem sizes for HACC simulations:
```
Small:
21000^2 global grid
1D FFT length: 21000 = 2^3 * 3^1 * 5^3 * 7^1
FFT batch size per MPI rank: 3675

Medium:
29400^3 global grid
1D FFT length: 29400 = 2^3 * 3^1 * 5^2 * 7^2
FFT batch size per MPI rank: 7203

Large:
38400^3 global grid
1D FFT length: 38400 = 2^9 * 3^1 * 5^2
FFT batch size per MPI rank: 12288
```
HACC often uses grid sizes with small primes up to 7. Use of grid sizes with primes large than 7 is not prohibited, but is used less often in production simulations.
